using System;
using System.IO;
using System.ComponentModel;

namespace id.Quake.Files {

	[SupportedExtension("cin")]
	public class Cinematic : INamedResource {

		#region Types

		private class HuffmanNode {
			public int Count;
			public bool Used;
			public int[] Children;
			public HuffmanNode() {
				this.Children = new int[2];
			}
		}

		#endregion

		#region Public Properties

		private readonly int width;
		/// <summary>Gets the width of the cinematic in pixels.</summary>
		[Category("Video")]
		[Description("Width of the video in pixels.")]
		public int Width { get { return this.width; } }

		private readonly int height;
		/// <summary>Gets the height of the cinematic in pixels.</summary>
		[Category("Video")]
		[Description("Height of the video in pixels.")]
		public int Height { get { return this.height; } }

		private readonly int sampleRate;
		/// <summary>Gets the audio sample rate.</summary>
		[Category("Audio")]
		[Description("Audio sample rate in Hz.")]
		public int SampleRate {
			get { return this.sampleRate; }
		}

		private readonly int sampleWidth;
		/// <summary>Gets the audio sample width (in bytes).</summary>
		[Category("Audio")]
		[Description("Width of the audio samples in bytes.")]
		public int SampleWidth {
			get { return this.sampleWidth; }
		}

		private readonly int channels;
		/// <summary>Gets the number of audio channels.</summary>
		[Category("Audio")]
		[Description("The number of available channels of audio.")]
		public int Channels {
			get { return this.channels; }
		}

		private bool isAtVideoEnd = false;
		/// <summary>Returns true if the cinematic is at the </summary>
		[Browsable(false)]
		public bool IsAtVideoEnd {
			get { return this.isAtVideoEnd; }
		}

		private string name;
		[Browsable(false)]
		public string Name { get { return this.name; } set { this.name = value; } }

		[Browsable(false)]
		public bool NamedInternally { get { return false; } }

		#endregion

		#region Private Member Variables

		/// <summary>Stream used to load data.</summary>
		private readonly Stream Data;

		/// <summary>Binary data tree used to decode the video frames.</summary>
		private readonly HuffmanNode[][] Nodes;

		/// <summary>???</summary>
		private int[] RootHuffmanNodeIndex;

		/// <summary>The working palette.</summary>
		private int[] Palette;

		/// <summary>Pointer to the a of the first frame of video data.</summary>
		private readonly long StartOfVideoData;


		private int FrameNumber;

		#endregion

		#region Public Methods

		/// <summary>Seek to the a of the cinematic.</summary>
		public void SeekToStart() {
			this.Palette = new int[256];
			this.Data.Seek(this.StartOfVideoData, SeekOrigin.Begin);
			this.isAtVideoEnd = false;
			this.FrameNumber = 0;
		}

		/// <summary>Get a frame's worth of data and moves on to the next frame.</summary>
		/// <returns>True if a frame was read, false otherwise.</returns>
		public bool ReadFrame(out Picture video, out byte[] sound) {

			BinaryReader FrameReader = new BinaryReader(this.Data);

			uint SequenceCommand = FrameReader.ReadUInt32();

			switch (SequenceCommand) {
				case 0x0000: // Nothing to do.
					break;
				case 0x0001: // Load the palette.
					for (int i = 0; i < 256; ++i) this.Palette[i] = unchecked((FrameReader.ReadByte() << 16) | (FrameReader.ReadByte() << 8) | (FrameReader.ReadByte() << 0) | (int)0xFF000000);
					break;
				case 0x0002: // Done!
					this.isAtVideoEnd = true;
					video = null;
					sound = null;
					return false;
			}


			uint CompressedSize = FrameReader.ReadUInt32();
			uint DecompressedSize = FrameReader.ReadUInt32();

			// Quick sanity check:
			if (DecompressedSize != this.Width * this.Height) throw new InvalidDataException("Frame is the wrong size.");

			int[] imageOut = new int[DecompressedSize];

			byte LastPixel = 0;
			byte WorkingValue = 0;
			int ImageOutIndex = 0;

			int BitPosition = 0;


			for (int i = 0; i < DecompressedSize; ++i) {

				int NodeIndex = this.RootHuffmanNodeIndex[LastPixel];

				HuffmanNode[] Nodes = this.Nodes[LastPixel];

				while (NodeIndex >= 256) {

					if (BitPosition == 0) {
						BitPosition = 8;
						WorkingValue = FrameReader.ReadByte();
					}

					NodeIndex = Nodes[NodeIndex].Children[WorkingValue & 0x01];
					WorkingValue >>= 1;
					--BitPosition;
				}
				imageOut[ImageOutIndex++] = this.Palette[LastPixel = (byte)NodeIndex];
			}

			
			int SoundSize = this.SampleWidth * this.Channels * this.SampleRate;
			int StartData = (FrameNumber + 0) * SoundSize / 14;
			int EndData = (FrameNumber + 1) * SoundSize / 14;
			sound = FrameReader.ReadBytes(EndData - StartData);

			++FrameNumber;

			video = new Picture(this.Width, this.Height, imageOut);
			return true;

		}

		#endregion

		#region Private Methods

		private void BuildHuffmanTree(int nodeIndex) {

			HuffmanNode[] WorkingNodes = this.Nodes[nodeIndex];

			for (int i = 0; i < 512; ++i) WorkingNodes[i].Used = false;

			int NodeCount = 256;

			for (; ; ) {
				HuffmanNode WorkingNode = WorkingNodes[NodeCount];
				WorkingNode.Children[0] = GetSmallestNode(WorkingNodes, NodeCount);
				if (WorkingNode.Children[0] == -1) break;
				WorkingNode.Children[1] = GetSmallestNode(WorkingNodes, NodeCount);
				if (WorkingNode.Children[1] == -1) break;

				WorkingNode.Count = WorkingNodes[WorkingNode.Children[0]].Count + WorkingNodes[WorkingNode.Children[1]].Count;

				++NodeCount;
			}

			this.RootHuffmanNodeIndex[nodeIndex] = NodeCount - 1;

		}

		private int GetSmallestNode(HuffmanNode[] nodes, int nodeCount) {

			int CurrentBestCount = int.MaxValue;
			int BestNode = -1;

			for (int i = 0; i < nodeCount; ++i) {

				if (nodes[i].Used) continue;
				if (nodes[i].Count == 0) continue;

				if (nodes[i].Count < CurrentBestCount) {
					CurrentBestCount = nodes[i].Count;
					BestNode = i;
				}

			}

			if (BestNode == -1) return -1;
			nodes[BestNode].Used = true;
			return BestNode;


		}

		#endregion

		#region Constructor

		public Cinematic(ResourceLoader resources, Stream stream) {

			this.Data = stream;
			BinaryReader CinematicReader = new BinaryReader(this.Data);

			// Header:
			this.width = CinematicReader.ReadInt32();
			this.height = CinematicReader.ReadInt32();
			this.sampleRate = CinematicReader.ReadInt32();
			this.sampleWidth = CinematicReader.ReadInt32();
			this.channels = CinematicReader.ReadInt32();

			// Huffman table:
			this.Nodes = new HuffmanNode[256][];
			for (int i = 0; i < 256; ++i) {
				this.Nodes[i] = new HuffmanNode[512];
				for (int j = 0; j < 512; ++j) this.Nodes[i][j] = new HuffmanNode();
			}
			this.RootHuffmanNodeIndex = new int[256];

			// Populate the nodes:
			for (int i = 0; i < 256; ++i) {
				for (int j = 0; j < 256; ++j) this.Nodes[i][j].Count = CinematicReader.ReadByte();
				this.BuildHuffmanTree(i);
			}

			// Initialise:
			this.StartOfVideoData = stream.Position;
			this.SeekToStart();

		}

		public void Dispose() {
			this.Data.Close();
			this.Data.Dispose();
		}

		#endregion

	}
}
