using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.ComponentModel;

namespace id.Quake.Files {
	public partial class Level {

		#region Properties

		private readonly Model[] models;
		/// <summary>Gets the list of models for the level.</summary>
		[Category("Geometry")]
		public Model[] Models {
			get { return this.models; }
		}

		#endregion

		/// <summary>Represents a single model.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public class Model : INamedResource, IBoundingBoxed {

			#region Public Properties

			private readonly BoundingBox bounds;
			/// <summary>Bounding box of the model.</summary>
			public BoundingBox Bounds { get { return this.bounds; } }

			private readonly Vector3 origin;
			/// <summary>Origin of the model.</summary>
			public Vector3 Origin { get { return this.origin; } }

			private readonly uint bspNodeIndex;
			/// <summary>Index of the first BSP node.</summary>
			public uint BspNodeIndex { get { return this.bspNodeIndex; } }

			private readonly uint clipNode0Index;
			/// <summary>Index of the first clip node.</summary>
			public uint ClipNode0Index { get { return this.clipNode0Index; } }

			private readonly uint clipNode1Index;
			/// <summary>Index of the second clip node.</summary>
			public uint ClipNode1Index { get { return this.clipNode1Index; } }

			private readonly uint node3;
			/// <summary>Usually zero.</summary>
			public uint Node3 { get { return this.node3; } }

			private readonly int leafCount;
			/// <summary>Number of BSP leaves.</summary>
			public int LeafCount { get { return this.leafCount; } }

			private readonly uint faceIndex;
			/// <summary>Index of the first face.</summary>
			public uint FaceIndex { get { return this.faceIndex; } }

			private readonly uint faceCount;
			/// <summary>Number of faces.</summary>
			public uint FaceCount { get { return this.faceCount; } }

			/// <summary>Returns a formatted string descibing the model.</summary>
			public override string ToString() {
				return this.Origin.ToString() + ", " + this.Bounds.ToString();
			}

			#endregion

			#region Private Members

			private readonly Level Level;

			#endregion

			#region Constructor

			public Model(Level level, Stream stream, int index) {

				this.Level = level;

				BinaryReader WorkingReader = new BinaryReader(stream);

				this.bounds = new BoundingBox(new Vector3(WorkingReader.BaseStream), new Vector3(WorkingReader.BaseStream));
				this.origin = new Vector3(WorkingReader.BaseStream);
				this.bspNodeIndex = WorkingReader.ReadUInt32();

				if (this.Level.FileVersion.GameVersion == Version.Game.Quake) {
					this.clipNode0Index = WorkingReader.ReadUInt32();
					this.clipNode1Index = WorkingReader.ReadUInt32();
					this.node3 = WorkingReader.ReadUInt32();
					this.leafCount = WorkingReader.ReadInt32();
				}

				this.faceIndex = WorkingReader.ReadUInt32();
				this.faceCount = WorkingReader.ReadUInt32();

				this.name = "*" + index;

				
			}

			#endregion

			#region INamedResource Members

			private string name;
			public string Name {
				get { return this.name; }
				set { this.name = value; }
			}

			public bool NamedInternally {
				get { return false; }
			}

			#endregion

			#region Public Methods

			/// <summary>Get the leaf that contains a point.</summary>
			/// <param name="point">The point to find a containing leaf for.</param>
			public Leaf GetLeafFromPoint(Vector3 point) {

				Node WorkingNode = this.Level.nodes[this.BspNodeIndex];

				while (WorkingNode != null) {
					// Are we in front of or behind the division?
					if (Vector3.Dot(this.Level.planes[WorkingNode.Plane].Normal, point) > this.Level.planes[WorkingNode.Plane].Offset) {
						if (WorkingNode.FrontIsNode) {
							WorkingNode = this.Level.nodes[WorkingNode.FrontIndex];
						} else {
							return this.Level.leaves[WorkingNode.FrontIndex];
						}
					} else {
						if (WorkingNode.BackIsNode) {
							WorkingNode = this.Level.nodes[WorkingNode.BackIndex];
						} else {
							return this.Level.leaves[WorkingNode.BackIndex];
						}
					}
				}

				return null;
			}

			/// <summary>Calculates the intersection point between a start position, end position, and the model.</summary>
			/// <param name="start">The position to start from.</param>
			/// <param name="end">The position to try to end at.</param>
			/// <returns>Information about the collision.</returns>
			public CollisionResult<Vector3> GetCollision(Vector3 start, Vector3 end) {
				return this.GetLeafFromPoint(start).GetCollision(start, end);	
			}

			#endregion

		}

	}
}
