using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.ComponentModel;

namespace id.Quake.Files {
	public partial class Level {

		#region Properties

		private readonly Plane[] planes;
		/// <summary>Gets the level's planes.</summary>
		[Category("Geometry")]	
		public Plane[] Planes {
			get { return this.planes; }
		}

		#endregion

		/// <summary>Describes a plane in 3D space.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public class Plane {

			#region Types

			/// <summary>The orientation of the plane.</summary>
			public enum PlaneAlignment {
				/// <summary>An axial plane in the X axis.</summary>
				AxialInX = 0,
				/// <summary>An axial plane in the Y axis.</summary>
				AxialInY = 1,
				/// <summary>An axial plane in the Z axis.</summary>
				AxialInZ = 2,
				/// <summary>A non-axial plane roughly aligned towards the X axis.</summary>
				NonAxialRoughlyX = 3,
				/// <summary>A non-axial plane roughly aligned towards the Y axis.</summary>
				NonAxialRoughlyY = 4,
				/// <summary>A non-axial plane roughly aligned towards the Z axis.</summary>
				NonAxialRoughlyZ = 5,
			}

			#endregion

			#region Properties

			private readonly Vector3 normal;
			/// <summary>Vector orthogonal to the plane.</summary>
			public Vector3 Normal {
				get { return this.normal; }
			}

			private readonly float offset;
			/// <summary>Offset to the plane from (0,0,0).</summary>
			public float Offset {
				get { return this.offset; }
			}

			private readonly PlaneAlignment alignment;
			/// <summary>Which plane the plane is most closely aligned to.</summary>
			public PlaneAlignment Alignment {
				get { return this.alignment; }
			}

			public override string ToString() {
				return this.normal.ToString() + " " + this.offset.ToString();
			}

			#endregion

			#region Constructor

			public Plane(Stream stream) {
				BinaryReader WorkingReader = new BinaryReader(stream);
				this.normal = new Vector3(stream);
				this.offset = WorkingReader.ReadSingle();
				this.alignment = (PlaneAlignment)WorkingReader.ReadUInt32();
			}

			#endregion

			#region Methods


			/// <summary>Calculate the intersection point of a point moving through the plane.</summary>
			/// <param name="start">The starting position of the point.</param>
			/// <param name="end">The desired end position of the point.</param>
			/// <returns>Either the end point or the intersection point, depending on whether the point hits the plane or not.</returns>
			/// <remarks>Optimised to take advantage of the alignment property.</remarks>
			public CollisionResult<Vector3> GetCollision(Vector3 start, Vector3 end) {

				switch (this.alignment) {
					case PlaneAlignment.AxialInX: {
							float IntersectionPoint = this.offset * this.normal.X;
							if (start.X != end.X && IntersectionPoint >= Math.Min(start.X, end.X) && IntersectionPoint <= Math.Max(start.X, end.X)) {
								float Distance = (IntersectionPoint - start.X) / (end.X - start.X);
								return new CollisionResult<Vector3>(
									true,
									new Vector3(
										IntersectionPoint,
										start.Y + (end.Y - start.Y) * Distance,
										start.Z + (end.Z - start.Z) * Distance
									), Distance, (Quake.Plane)this
								);
							} else {
								return new CollisionResult<Vector3>(false, end);
							}
						}
					case PlaneAlignment.AxialInY: {
							float IntersectionPoint = this.offset * this.normal.Y;
							if (start.Y != end.Y && IntersectionPoint >= Math.Min(start.Y, end.Y) && IntersectionPoint <= Math.Max(start.Y, end.Y)) {
								float Distance = (IntersectionPoint - start.Y) / (end.Y - start.Y);
								return new CollisionResult<Vector3>(
									true,
									new Vector3(
										start.X + (end.X - start.X) * Distance,
										IntersectionPoint,
										start.Z + (end.Z - start.Z) * Distance
									), Distance, (Quake.Plane)this
								);
							} else {
								return new CollisionResult<Vector3>(false, end);
							}
						}
					case PlaneAlignment.AxialInZ: {
							float IntersectionPoint = this.offset * this.normal.Z;
							if (start.Z != end.Z && IntersectionPoint >= Math.Min(start.Z, end.Z) && IntersectionPoint <= Math.Max(start.Z, end.Z)) {
								float Distance = (IntersectionPoint - start.Z) / (end.Z - start.Z);
								return new CollisionResult<Vector3>(
									true,
									new Vector3(
										start.X + (end.X - start.X) * Distance,
										start.Y + (end.Y - start.Y) * Distance,
										IntersectionPoint
									), Distance, (Quake.Plane)this
								);
							} else {
								return new CollisionResult<Vector3>(false, end);
							}
						}
					default:
						Quake.Plane P = (Quake.Plane)this;
						return P.GetCollision(start, end);
				}
			}


			/// <summary>Get the collision position between a bounding box and a plane.</summary>
			/// <param name="box">The bounding box to move.</param>
			/// <param name="direction">The direction vector to move the box by.</param>
			public CollisionResult<BoundingBox> GetCollision(BoundingBox box, Vector3 direction) {

				// Find the end position:
				BoundingBox End = box + direction;

				// Quick bounding-box intersection test:
				if (!((Quake.Plane)this).Intersects(BoundingBox.Combine(box, End))) {
					return new CollisionResult<BoundingBox>(false, End);
				}

				//OPTIMISATION: Axial planes.

				Quake.Plane P = (Quake.Plane)this;
				return P.GetCollision(box, direction);

			}


			#endregion

		}	
	}
}