using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.ComponentModel;

namespace id.Quake.Files {
	public partial class Level {


		#region Properties

		private readonly TextureInfo[] textureInformation;
		/// <summary>Gets the level's texture information.</summary>
		[Category("Textures")]
		public TextureInfo[] TextureInformation {
			get { return this.textureInformation; }
		}


		#endregion

		/// <summary>Represents information about how a texture is attached to a face.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public class TextureInfo : IPicture, IMipMappedPicture {

			private readonly Level Level;

			#region Types

			/// <summary>Describes special effects relating to the texture.</summary>
			/// <remarks>Used in Quake 2.</remarks>
			[Flags()] public enum SurfaceType {
				None = 0x00,
				/// <summary>Value will hold the light strength.</summary>
				Light = 0x01,
				/// <summary>Affects the game physics.</summary>
				Slick = 0x02,
				/// <summary>Don't draw, let skybox show through.</summary>
				Sky = 0x04,
				/// <summary>Turbulent water warp.</summary>
				Warp = 0x08,
				Translucent33 = 0x10,
				Translucent66 = 0x20,
				/// <summary>Scroll towards angle.</summary>
				Flowing = 0x40,
				/// <summary>Don't bother referencing this texture.</summary>
				NoDraw = 0x80,
			}

			#endregion

			#region Standard Properties

			private readonly Vector3 horizontal;
			/// <summary>Horizontal direction in texture space.</summary>
			public Vector3 Horizontal {
				get { return this.horizontal; }
			}

			private readonly float horizontalOffset;
			/// <summary>Horizontal offset in texture space.</summary>
			public float HorizontalOffset {
				get { return this.horizontalOffset; }
			}

			private readonly Vector3 vertical;
			/// <summary>Vertical direction in texture space.</summary>
			public Vector3 Vertical {
				get { return this.vertical; }
			}

			private readonly float verticalOffset;
			/// <summary>Vertical offset in texture space.</summary>
			public float VerticalOffset {
				get { return this.verticalOffset; }
			}


			private readonly uint textureId;
			/// <summary>Index of the texture itself.</summary>
			/// <remarks>Quake 1 only.</remarks>
			public uint TextureId {
				get { return this.textureId; }
			}

			private readonly bool isAnimated;
			/// <summary>Specifies whether the texture has a warped animation.</summary>
			/// <remarks>Quake 1 only.</remarks>
			public bool IsAnimated {
				get { return this.isAnimated; }
			}

			private readonly string name;
			/// <summary>Name of the texture.</summary>
			public string Name {
				get {
					switch (this.Level.FileVersion.GameVersion) {
						case Version.Game.Quake:
							return this.Level.MipTextures[this.TextureId].Name;
						case Version.Game.Quake2:
							return this.name; 
						default:
							throw new InvalidOperationException();
					}
				}
			}

			private readonly SurfaceType flags;
			/// <summary>Texture flags.</summary>
			/// <remarks>Introduced in Quake 2.</remarks>
			public SurfaceType Flags {
				get { return this.flags; }
			}

			private readonly uint value;
			/// <summary>Texture value.</summary>
			/// <remarks>Introduced in Quake 2.</remarks>
			public uint Value {
				get { return this.value; }
			}
			#endregion

			#region Extended Properties

			/// <summary>Returns true if the texture is a sky texture.</summary>
			public bool IsSky {
				get {
					switch (this.Level.FileVersion.GameVersion) {
						case Version.Game.Quake:
							return this.Level.MipTextures[this.TextureId].Name.ToLower().StartsWith("sky");
						case Version.Game.Quake2:
							return (this.Flags & SurfaceType.Sky) != SurfaceType.None;
						default:
							throw new InvalidOperationException();
					}
				}
			}

			/// <summary>Returns true if the texture is a water, slime, lava or similar texture.</summary>
			public bool IsLiquid {
				get {
					switch (this.Level.FileVersion.GameVersion) {
						case Version.Game.Quake:
							return this.IsAnimated;
						case Version.Game.Quake2:
							return (this.Flags & SurfaceType.Warp) != SurfaceType.None;
						default:
							throw new InvalidOperationException();
					}
				}
			}

			/// <summary>Returns true if the texture is visible.</summary>
			/// <remarks>Textures that aren't drawn directly, such as the skybox textures, still return true.</remarks>
			public bool IsVisible {
				get {
					switch (this.Level.FileVersion.GameVersion) {
						case Version.Game.Quake:
							return this.Name != "trigger";
						case Version.Game.Quake2:
							return (this.Flags & SurfaceType.NoDraw) == SurfaceType.None;
						default:
							throw new InvalidOperationException();
					}
				}
			}

			/// <summary>Returns the name of the next frame's texture if this is an animated texture.</summary>
			public string NextName {
				get {
					switch (this.Level.FileVersion.GameVersion) {
						case Version.Game.Quake:
							if (this.Name.Length >= 2 && this.Name[0] == '+' && this.Name[1] >= '0' && this.Name[1] <= '9') {
								int FrameNumber = ((this.Name[1] - '0') + 1) % 10;
								string NextFrameName = "+" + FrameNumber.ToString() + this.Name.Substring(2);
								for (int i = 0; i < this.Level.textures.Length; ++i) if (this.Level.textures[i].Name == NextFrameName) return NextFrameName;
								NextFrameName = "+0" + this.Name.Substring(2);
								for (int i = 0; i < this.Level.textures.Length; ++i) if (this.Level.textures[i].Name == NextFrameName) return NextFrameName;
								return "";
							} else {
								return "";
							}
						case Version.Game.Quake2:
							return this.Level.Resources.LoadObject<WallTexture>("textures/" + this.Name + ".wal").NextName;
						default:
							throw new InvalidOperationException();
					}
				}
			}

			public override string ToString() {
				return this.Name;
			}

			#endregion

			#region Methods

			/// <summary>Calculate the conventional texture coordinates for a vertex on a face using this texture information.</summary>
			/// <param name="level">The level the vertex and texture information comes from.</param>
			/// <param name="normalise">Whether to normalise the coordinate between 0 and 1.</param>
			/// <param name="vertex">The vertex to calculate texture coordinates for.</param>
			public Vector2 GetTextureCoordinates(Vector3 vertex, bool normalise) {
				Vector2 Coordinate = new Vector2(
					(Vector3.Dot(vertex, this.Horizontal) + this.HorizontalOffset),
					(Vector3.Dot(vertex, this.Vertical) + this.VerticalOffset)
				);
				if (!normalise) return Coordinate;
				Picture P = this.Picture;
				return Coordinate / new Vector2((float)P.Width, (float)P.Height);
			}

			/// <summary>Calculate the normalised texture coordinates for a vertex on a face using this texture information.</summary>
			/// <param name="vertex">The vertex to calculate texture coordinates for.</param>
			public Vector2 GetTextureCoordinates(Vector3 vertex) {
				return this.GetTextureCoordinates(vertex, true);
			}

	
			#endregion

			#region Constructor

			public TextureInfo(Level level, Stream stream) {
				this.Level = level;

				BinaryReader WorkingReader = new BinaryReader(stream);

				this.horizontal = new Vector3(WorkingReader.BaseStream);
				this.horizontalOffset = WorkingReader.ReadSingle();
				this.vertical = new Vector3(WorkingReader.BaseStream);
				this.verticalOffset = WorkingReader.ReadSingle();


				switch (level.FileVersion.GameVersion) {
					case Version.Game.Quake:
						this.textureId = WorkingReader.ReadUInt32();
						this.isAnimated = WorkingReader.ReadUInt32() != 0;
						break;
					case Version.Game.Quake2:
						this.flags = (SurfaceType)WorkingReader.ReadInt32();
						this.value = WorkingReader.ReadUInt32();
						this.name = Encoding.ASCII.GetString(WorkingReader.ReadBytes(32)).Split('\0')[0];
						uint NextTexture = WorkingReader.ReadUInt32();
						break;
					default:
						throw new InvalidOperationException();
				}
			}

			#endregion


			#region IPicture Members

			private Picture CachedPicture = null;

			public Picture Picture {
				get {
					if (this.CachedPicture == null) {
						if (this.Level.FileVersion.SupportsEntry(Entry.EntryTypes.MipTextures)) {
							this.CachedPicture = this.Level.MipTextures[this.TextureId].Picture;
						} else {
							this.CachedPicture = new WallTexture(this.Level.Resources, this.Level.Resources.OpenFile("textures/" + this.Name + ".wal")).Picture;
						}
					}
					return this.CachedPicture;
				}
			}

			#endregion

			public Picture GetPicture(int mipLevel) {
				if (this.Level.FileVersion.SupportsEntry(Entry.EntryTypes.MipTextures)) {
					return this.Level.MipTextures[this.TextureId].GetPicture(mipLevel);
				} else {
					return this.Level.Resources.LoadObject<WallTexture>("textures/" + this.Name + ".wal").GetPicture(mipLevel);
				}
			}

		}
	}
}