using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;

namespace id.Quake {

	/// <summary>Defines a simple bounding box.</summary>
	[TypeConverter(typeof(ExpandableObjectConverter))]
	public struct BoundingBox {

		/// <summary>Gets or sets the minimum position in the bounding box.</summary>
		public Vector3 Min;
		/// <summary>Gets or sets the maximum position in the bounding box.</summary>
		public Vector3 Max;

		/// <summary>Creates an instance of the bounding box from two positions.</summary>
		/// <param name="min">The smallest possible position in the bounding box.</param>
		/// <param name="max">The largest possible position in the bounding box.</param>
		public BoundingBox(Vector3 min, Vector3 max) {
			this.Min = min;
			this.Max = max;
		}


		/// <summary>Returns true if the bounding box is valid, false otherwise.</summary>
		public bool IsValid {
			get { return Min.X <= Max.X && Min.Y <= Max.Y && Min.Z <= Max.Z; }
		}

		/// <summary>Check if a bounding box completely contains another.</summary>
		public bool CompletelyContains(BoundingBox b) {
			return
				(b.Min.X >= this.Min.X) && (b.Max.X <= this.Max.X) &&
				(b.Min.Y >= this.Min.Y) && (b.Max.Y <= this.Max.Y) &&
				(b.Min.Z >= this.Min.Z) && (b.Max.Z <= this.Max.Z);
		}

		/// <summary>Check if another bounding box intersects with this one.</summary>
		/// <param name="b">The bounding box to check.</param>
		public bool Intersects(BoundingBox b) {
			return !(
				b.Min.X > this.Max.X || b.Max.X < this.Min.X ||
				b.Min.Y > this.Max.Y || b.Max.Y < this.Min.Y ||
				b.Min.Z > this.Max.Z || b.Max.Z < this.Min.Z
			);
		}

		/// <summary>Combine two bounding boxes together to get a bounding box that completely contains both.</summary>
		public static BoundingBox Combine(BoundingBox a, BoundingBox b) {
			return new BoundingBox(Vector3.Min(a.Min, b.Min), Vector3.Max(a.Max, b.Max));
		}

		/// <summary>Add a vector to a bounding box to move it.</summary>
		public static BoundingBox operator +(BoundingBox box, Vector3 direction) {
			return new BoundingBox(box.Min + direction, box.Max + direction);
		}

		/// <summary>Returns a formatted string describing the bounding box.</summary>
		public override string ToString() {
			return this.Min.ToString() + "->" + this.Max.ToString();
		}

		/// <summary>Gets the coordinates of the point that lies in the centre of the bounding box.</summary>
		public Vector3 MidPoint {
			get { return (Min + Max) * 0.5f; }
		}

		/// <summary>Gets an array of <c>Vector3</c>s, one for each corner of the bounding box.</summary>
		public Vector3[] Corners {
			get {
				return new Vector3[]{
					new Vector3(Min.X, Min.Y, Min.Z),
					new Vector3(Min.X, Min.Y, Max.Z),
					new Vector3(Min.X, Max.Y, Min.Z),
					new Vector3(Min.X, Max.Y, Max.Z),
					new Vector3(Max.X, Min.Y, Min.Z),
					new Vector3(Max.X, Min.Y, Max.Z),
					new Vector3(Max.X, Max.Y, Min.Z),
					new Vector3(Max.X, Max.Y, Max.Z)
				};
			}
		}
	}
}
