using System;
using System.Collections.Generic;
using QuakeC = id.Quake.Files.QuakeC;

namespace id.Quake {
	public partial class VirtualMachine {

		#region Private Members

		/// <summary>Keep track of the current working function.</summary>
		private QuakeC.Function WorkingFunction;

		#endregion

		#region Public Invoke Methods

		/// <summary>Invoke a QuakeC function.</summary>
		/// <param name="function">The function to invoke.</param>
		public void Invoke(QuakeC.Function function) {
			this.ExecuteFunctionBody(function);
		}

		/// <summary>Invoke a QuakeC function.</summary>
		/// <param name="function">The name of the function to invoke.</param>
		public void Invoke(string function) {
			this.Invoke(this.quakeCSource.Functions[function]);
		}

		/// <summary>Invoke a QuakeC function on an entity.</summary>
		/// <param name="edict">The entity to invoke the function on.</param>
		/// <param name="function">The function to invoke.</param>
		/// <remarks>Modifies the global 'self' variable.</remarks>
		public void Invoke(Edict edict, QuakeC.Function function) {
			this.Self.Value = edict;
			this.Invoke(function);
		}

		/// <summary>Invoke a QuakeC function on an entity.</summary>
		/// <param name="edict">The entity to invoke the function on.</param>
		/// <param name="function">The name of the function to invoke.</param>
		/// <remarks>Modifies the global 'self' variable.</remarks>
		public void Invoke(Edict edict, string function) {
			this.Self.Value = edict;
			this.Invoke(function);
		}

		/// <summary>Invoke main().</summary>
		public void InvokeMain() {
			this.Invoke((QuakeC.Function)this.quakeCSource.Globals["main"].Value);
		}

		#endregion

		#region Stacks

		/// <summary>Used to store a function and statement pointer on the stack.</summary>
		private struct StackEntry {
			public int StatementPointer;
			public QuakeC.Function Function;
			public StackEntry(int statementPointer, QuakeC.Function function) {
				this.StatementPointer = statementPointer;
				this.Function = function;
			}
		}

		/// <summary>Stack used to store functions and stack pointers for function calling.</summary>
		private Stack<StackEntry> FunctionStack;

		/// <summary>Used to store temporary local variables.</summary>
		private Stack<int> LocalStack;

		#endregion

		#region Flow Control
		
		/// <summary>Enter a QuakeC function.</summary>
		/// <param name="function">The function to enter.</param>
		/// <remarks>Stores local values onto the stack and loads parameters.</remarks>
		private void EnterFunction(QuakeC.Function function) {

			// Preserve locals:
			for (int i = 0, p = function.FirstParameter; i < function.Locals; ++i) {
				LocalStack.Push(this.quakeCSource.GlobalData[p].Integer);
				++p;
			}

			// Copy parameters:
			for (int i = 0, p = function.FirstParameter; i < function.ParameterCount; ++i) {
				for (int j = 0; j < function.ParameterSizes[i]; ++j) {
					this.quakeCSource.GlobalData[p].Integer = this.quakeCSource.GlobalData[(int)WorkingVariableOffsets.Parameter0 + i * 3 + j].Integer;
					++p;
				}
			}

			this.FunctionStack.Push(new StackEntry(this.StatementPointer, this.WorkingFunction));
			this.StatementPointer = function.FirstStatement;
			this.WorkingFunction = function;
		}

		/// <summary>Leave a QuakeC function.</summary>
		/// <remarks>Restores local values from the stack.</remarks>
		private void LeaveFunction() {

			// Restore locals:
			for (int i = 0, p = this.WorkingFunction.FirstParameter + this.WorkingFunction.Locals; i < this.WorkingFunction.Locals; ++i) {
				--p;
				this.quakeCSource.GlobalData[p].Integer = LocalStack.Pop();
			}
			StackEntry SE = this.FunctionStack.Pop();
			this.WorkingFunction = SE.Function;
			this.StatementPointer = SE.StatementPointer;
		}

		#endregion

		#region Setup

		/// <summary>Initialise function-related stuff.</summary>
		private void InitialiseFunctions() {
			this.FunctionStack = new Stack<StackEntry>(256);
			this.LocalStack = new Stack<int>(256);
			this.WorkingFunction = null;
		}

		#endregion


	}
}
