using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using System.IO;
using System.Reflection;

namespace Ranger {
	public partial class RangerFrontEnd : Form {

		private id.Quake.ResourceLoader Resources;


		private void SetDirectory(string path) {
			
			this.Resources = null;
			this.AllFiles = new string[0];
			RefreshFolders();
			this.FileList.Items.Clear();

			this.Resources = new id.Quake.ResourceLoader(path);
			this.Resources.AddStandardPackages();
			this.AllFiles = this.Resources.GetFiles();
			this.RefreshFolders();

			Properties.Settings.Default.LastPath = path;
			Properties.Settings.Default.Save();

		}

		private Type[] Loaders = new Type[] { 
			typeof(id.Quake.Files.Cinematic),
			typeof(id.Quake.Files.WallTexture),
			typeof(id.Quake.Files.PictureLump),
			typeof(id.Quake.Files.ZSoftPcx),
			typeof(id.Quake.Files.Level),
			typeof(id.Quake.Files.QuakeC),
			typeof(id.Quake.Files.ColourMap),
			typeof(id.Quake.Files.Palette),
			typeof(id.Quake.Files.Configuration),
			typeof(id.Quake.Files.Model),
		};

		private string[] AllFiles;

		public RangerFrontEnd() {

			InitializeComponent();
			this.Text = Application.ProductName;

			this.InterfaceIcons.Images.Add(Properties.Resources.Icon_Folder);
			this.InterfaceIcons.Images.Add(Properties.Resources.Icon_Drive);
			this.InterfaceIcons.Images.Add(Properties.Resources.Icon_Page);

			if (Directory.Exists(Properties.Settings.Default.LastPath)) {
				this.SetDirectory(Properties.Settings.Default.LastPath);
			}

		}



		private void RefreshFolders() {
			List<string> AddedFolders = new List<string>(this.AllFiles.Length / 4);

			foreach (string s in this.AllFiles) {
				string t = "/" + Path.GetDirectoryName(s).Replace('\\', '/');
				if (!AddedFolders.Contains(t)) AddedFolders.Add(t);
				string[] Components = t.Split('/');

				string TestComponentExists = "/" + Components[1];
				for (int i = 2; i < Components.Length; ++i) {
					if (!AddedFolders.Contains(TestComponentExists)) AddedFolders.Add(TestComponentExists);
					TestComponentExists += "/" + Components[i];
				}
			}

			AddedFolders.Sort();

			TreeNode RootNode = new TreeNode();

			if (this.Resources != null) {
				PopulateTreeNodeFolders(RootNode, "/", AddedFolders.ToArray());
				RootNode.Text = Path.GetFileName(this.Resources.BaseDirectory);
				this.FolderList.Nodes.Clear();
				this.FolderList.Nodes.Add(RootNode);
				RootNode.Expand();
			} else {
				this.FolderList.Nodes.Clear();
			}
		}

		private void PopulateTreeNodeFolders(TreeNode node, string baseDirectory, string[] allFolders) {
			node.Text = Path.GetFileName(baseDirectory);
			node.ImageIndex = 0;
			node.SelectedImageIndex = 0;
			node.Tag = baseDirectory;

			foreach (string s in allFolders) {
				if (s.StartsWith(baseDirectory) && s.Length > baseDirectory.Length) {
					string[] PartsAfter = s.Substring(baseDirectory.Length + 1).Split('/');
					if (PartsAfter.Length == 1 && PartsAfter[0] != "") {
						TreeNode SubNode = new TreeNode();
						node.Nodes.Add(SubNode);
						PopulateTreeNodeFolders(SubNode, s, allFolders);
					}
				}
			}


		}

		private void FolderList_AfterSelect(object sender, TreeViewEventArgs e) {

			List<ListViewItem> FilesToList = new List<ListViewItem>(this.AllFiles.Length / 8);
			if (e.Node != null) {

				string Directory = (e.Node as TreeNode).Tag as string;
				string BaseDirectory = (Directory).ToLowerInvariant();
				this.SelectedFilePath.Text = Directory;

				foreach (string s in this.AllFiles) {
					if ("/" + Path.GetDirectoryName(s).Replace('\\', '/').ToLowerInvariant() == BaseDirectory) {
						ListViewItem LVI = new ListViewItem(Path.GetFileName(s));

						if (Resources.FileIsPackaged(s)) {
							LVI.ImageIndex = 2;
						} else {
							LVI.ImageIndex = 1;
						}

						LVI.Tag = s;
						FilesToList.Add(LVI);
					}
				}

			} else {
				this.SelectedFilePath.Text = null;
			}

			FilesToList.Sort(delegate(ListViewItem a, ListViewItem b) { return a.Text.CompareTo(b.Text); });

			this.FileList.Items.Clear();
			this.FileList.Items.AddRange(FilesToList.ToArray());
		}

		private void FileList_ItemActivate(object sender, EventArgs e) {
			if (FileList.SelectedItems.Count == 1) PreviewObject(this.PropertyViewer.SelectedObject);
			
		}

		private void PreviewObject(object o) {
			if (o != null) {

				string Filename = FileList.SelectedItems.Count == 1 ? Path.GetFileName(FileList.SelectedItems[0].Tag as string) : "Unknown";

				if (o.GetType() == typeof(id.Quake.Files.Cinematic)) {
					new CinematicPreviewer(Filename, o as id.Quake.Files.Cinematic).Show(this);
					return;
				} else if (o.GetType() == typeof(id.Quake.Files.Level)) {
					new MapPreviewer(Filename, o as id.Quake.Files.Level).Show(this);
					return;
				}

				foreach (Type t in o.GetType().GetInterfaces()) {

					if (t == typeof(id.Quake.Files.IPicture)) {

						new PicturePreviewer(
								Path.GetFileName(Filename),
								(o as id.Quake.Files.IPicture).Picture
							).Show(this);

					}

					if (t == typeof(id.Quake.Files.IText)) {
						new TextPreviewer(
								Path.GetFileName(Filename),
								(o as id.Quake.Files.IText).Text
							).Show(this);
					}

				}
			}
		}

		private object LoadResource(string name) {
			string Extension = Path.GetExtension(name);
			foreach (Type t in Loaders) {
				object[] SupportedExtensions = t.GetCustomAttributes(typeof(id.Quake.Files.SupportedExtension), false);
				foreach (object o in SupportedExtensions) {
					if (("." + (o as id.Quake.Files.SupportedExtension).Extension) == Extension) {
						ConstructorInfo C = t.GetConstructor(new Type[] { typeof(id.Quake.ResourceLoader), typeof(Stream) });
						if (C == null) continue;
						try {
							return C.Invoke(new object[] { this.Resources, this.Resources.OpenFile(name) });
						} catch (Exception e) {
							MessageBox.Show(this, "Error decoding file: " + (e.InnerException != null ? e.InnerException.Message : e.Message), name, MessageBoxButtons.OK, MessageBoxIcon.Error);
						}
					}
				}
			}
			return null;
		}

		private void FileList_SelectedIndexChanged(object sender, EventArgs e) {


			if (FileList.SelectedItems.Count == 1) {
				string Filename = FileList.SelectedItems[0].Tag as string;
				this.PropertyViewer.SelectedObject = LoadResource(Filename);
				this.SelectedFilePath.Text = "/" + Filename;
			} else {
				this.PropertyViewer.SelectedObject = null;
				this.SelectedFilePath.Text = null;
			}
		}

		private void MenuFileExit_Click(object sender, EventArgs e) {
			this.Close();
		}

		private void MenuFileSetBase_Click(object sender, EventArgs e) {
			using (FolderBrowserDialog FBD = new FolderBrowserDialog()) {
				if (FBD.ShowDialog(this) == DialogResult.OK) {
					this.SetDirectory(FBD.SelectedPath);
				}
			}
		}

		private void MenuFile_DropDownOpening(object sender, EventArgs e) {
			this.MenuFileExport.Enabled = this.FileList.SelectedItems.Count > 0;
		}

		private void MenuFileExport_Click(object sender, EventArgs e) {

			string DestinationPath = "";
			if (this.FileList.SelectedItems.Count == 1) {
				string FileToSave = this.FileList.SelectedItems[0].Tag as string;
				this.ExportSave.Filter = "Quake File (*" + Path.GetExtension(FileToSave) + ")|*" + Path.GetExtension(FileToSave);
				this.ExportSave.FileName = Path.GetFileName(FileToSave);
				if (this.ExportSave.ShowDialog(this) != DialogResult.OK) return;
				DestinationPath = this.ExportSave.FileName;
			} else {
				if (this.ExportSaveFolder.ShowDialog(this) != DialogResult.OK) return;
				DestinationPath = this.ExportSaveFolder.SelectedPath;
			}


			foreach (ListViewItem ToExport in this.FileList.SelectedItems) {
				using (BinaryReader SourceReader = new BinaryReader(this.Resources.OpenFile(ToExport.Tag as string))) {
					string DestinationName = this.FileList.SelectedItems.Count == 1 ? DestinationPath : Path.Combine(DestinationPath, Path.GetFileName(ToExport.Tag as string));
					
					// Check overwrite:
					if (File.Exists(DestinationName)) {
						switch (MessageBox.Show(this, "The file '" + Path.GetFileName(DestinationName) + "' already exists." + Environment.NewLine + Environment.NewLine + "Would you like to overwrite it?", "Export", MessageBoxButtons.YesNoCancel)) {
							case DialogResult.Yes:
								break;
							case DialogResult.Cancel:
								return;
							default:
								continue;
						}
					}

					File.WriteAllBytes(DestinationName, SourceReader.ReadBytes((int)SourceReader.BaseStream.Length));

				}
			}
		}

		private void PropertyGridContext_Opening(object sender, CancelEventArgs e) {
			e.Cancel = this.PropertyViewer.SelectedGridItem == null;
		}

		private void PreviewPropertyGridItem_Click(object sender, EventArgs e) {
			PreviewObject(this.PropertyViewer.SelectedGridItem.Value);
		}


	}
}