using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;

namespace XnaQuake.Renderer {
	class RenderController : ILoadable {

		#region Public Members / Properties

		public VectorRendererComponent.VectorRenderer VectorRenderer;

		/// <summary>The shader effect used for rendering.</summary>
		public Effect WorldEffect;

		/// <summary>The graphics device manager for the rendering.</summary>
		public GraphicsDeviceManager Graphics;
		
		/// <summary>View matrix.</summary>
		public Matrix View;

		/// <summary>Projection matrix.</summary>
		public Matrix Projection;

		/// <summary>World matrix.</summary>
		public Matrix World;

		public Matrix BaseWorldViewProj;

		public Matrix FishEyeCamera;

		/// <summary>Gets the current frame number.</summary>
		public int FrameNumber;

		/// <summary>Gets the light styles.</summary>
		public float[][] LightStyles;
		public float[] CurrentLightStyleLevels;
		public int[] LightStyleFrames;

		private GameTime currentTime;
		public GameTime Time {
			get { return this.currentTime; }
		}

		#region Camera


		private Vector3 cameraAngle;
		public Vector3 CameraAngle {
			get { return this.cameraAngle; }
			set { this.cameraAngle = value; this.UpdateCamera(); }
		}

		private Vector3 cameraPosition;
		public Vector3 CameraPosition {
			get { return this.cameraPosition; }
			set { this.cameraPosition = value; this.UpdateCamera(); }
		}

		private void UpdateCamera() {
			Matrix YPR = Matrix.CreateFromYawPitchRoll(this.cameraAngle.X, this.cameraAngle.Y, this.cameraAngle.Z);

			this.View = Matrix.CreateLookAt(
				Vector3.Zero,
				Vector3.Transform(Vector3.Forward, YPR),
				Vector3.Up);
			
			this.World = Matrix.CreateTranslation(-this.cameraPosition);
		}

		#endregion

		#endregion

		#region Private Members / Properties

		private VertexDeclaration VertexDeclaration;

		#endregion

		#region Constructor

		public RenderController() {
			this.LightStyles = new float[64][];
			this.LightStyleFrames = new int[64];
			this.CurrentLightStyleLevels = new float[64];
			for (int i = 0; i < LightStyles.Length; ++i) {
				LightStyles[i] = new float[] { 0.5f };
				this.CurrentLightStyleLevels[i] = 0.5f;
			}
		}

		#endregion

		double LightTimeElapsed = 0;
		public void Draw(GameTime time, IDrawable[] toDraw) {
			
			this.currentTime = time;

			++FrameNumber;

			LightTimeElapsed += time.ElapsedGameTime.TotalSeconds;
			while (LightTimeElapsed > 0) {
				for (int i = 0; i < 64; ++i) {
					LightStyleFrames[i] = (LightStyleFrames[i] + 1) % LightStyles[i].Length;
					CurrentLightStyleLevels[i] = LightStyles[i][LightStyleFrames[i]];
				}
				LightTimeElapsed -= 0.1d;
			}

			this.Graphics.GraphicsDevice.Clear(Color.Black);
			
			this.Graphics.GraphicsDevice.RenderState.DepthBufferEnable = true;
			this.Graphics.GraphicsDevice.RenderState.AlphaBlendEnable = false;

			this.Graphics.GraphicsDevice.VertexDeclaration = this.VertexDeclaration;

			this.BaseWorldViewProj = this.World * this.View * this.Projection;
			this.WorldEffect.Parameters["WorldViewProj"].SetValue(this.BaseWorldViewProj);
			this.WorldEffect.Parameters["ViewProj"].SetValue(this.View * this.Projection);

			this.WorldEffect.Parameters["Time"].SetValue((float)DateTime.Now.TimeOfDay.TotalSeconds);

			foreach (IDrawable ToDraw in toDraw) {
				ToDraw.Draw(this);				
			}
		}

		#region ILoadable Members

		private Texture2D ColourMap;

		public void Load(id.Quake.ResourceLoader resources, GraphicsDeviceManager graphics, ContentManager content) {

			this.WorldEffect = content.Load<Effect>("World");

			id.Quake.Files.IPicture ColourMapLoader = (resources.FileExists("gfx/colormap.lmp")) ?
				(id.Quake.Files.IPicture)resources.LoadObject<id.Quake.Files.ColourMap>("gfx/colormap.lmp") :
				(id.Quake.Files.IPicture)resources.LoadObject<id.Quake.Files.ZSoftPcx>("pics/colormap.pcx");

			id.Quake.Picture Picture = ColourMapLoader.Picture;
			if (Picture.Width != 256 || Picture.Height < 64) throw new System.IO.InvalidDataException("Colour map is the wrong size.");

			int[] CroppedPicture = new int[256 * 64];
			Array.Copy(Picture.Data, CroppedPicture, CroppedPicture.Length);
			this.ColourMap = Utility.TextureFromPicture(graphics.GraphicsDevice, new id.Quake.Picture(256, 64, CroppedPicture));

			this.WorldEffect.Parameters["ColourMap"].SetValue(ColourMap);

			this.Graphics = graphics;

			// Set up the projection matrix:
			this.UpdateProjection();

			// ...and view:
			this.CameraAngle = Vector3.Zero;
			this.CameraPosition = Vector3.Zero;

			// Vertex declaration:
			this.VertexType = BspRenderer.VertexPositionTextureDiffuseLightMap.VertexElements;

		}


		public void UpdateProjection() {
			this.Projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.ToRadians(90), (float)Graphics.GraphicsDevice.Viewport.Width / (float)Graphics.GraphicsDevice.Viewport.Height, 1.0f, 5000.0f);
		}

		public VertexElement[] VertexType {
			set {
				this.VertexDeclaration = new VertexDeclaration(Graphics.GraphicsDevice, value);
				this.Graphics.GraphicsDevice.VertexDeclaration = this.VertexDeclaration;
			}
		}

		#endregion

		#region ILoadable Members


		public void Unload(GraphicsDeviceManager graphics, ContentManager content) {
			this.ColourMap.Dispose();
			this.WorldEffect.Dispose();
		}

		#endregion
	}
}
